@@ -1,3 +1,17 @@
+/*
+	$Id: comio.h,v 1.2 1998/10/18 13:18:27 crosser Exp $
+*/
+
+/*
+	$Log: comio.h,v $
+	Revision 1.2  1998/10/18 13:18:27  crosser
+	Put RCS logs and I.D. into the source
+
+	Revision 1.1  1998/01/18 02:16:45  crosser
+	Initial revision
+	
+*/
+
 #ifndef _COMIO_H
 #define _COMIO_H
 
