@@ -1,9 +1,9 @@
 #ifndef LINT
-static char *rcsid="$Id: photopc.c,v 2.74 2000/02/17 21:29:55 crosser Exp $";
+static char *rcsid="$Id: photopc.c,v 2.83 2001/01/12 23:26:40 crosser Exp $";
 #endif
 
 /*
-	Copyright (c) 1997-1999 Eugene G. Crosser
+	Copyright (c) 1997-2001 Eugene G. Crosser
 	Copyright (c) 1998,1999 Bruce D. Lightner (DOS/Windows support)
 
 	You may distribute and/or use for any purpose modified or unmodified
@@ -17,6 +17,33 @@ static char *rcsid="$Id: photopc.c,v 2.74 2000/02/17 21:29:55 crosser Exp $";
 
 /*
 	$Log: photopc.c,v $
+	Revision 2.83  2001/01/12 23:26:40  crosser
+	new release, avoid strncasecmp, change copyright year
+	
+	Revision 2.82  2001/01/12 11:07:03  crosser
+	If root directory is non-empty, skip it anyway because buggy models
+	report non-zero number of pictures even when current directory is root.
+	Also small fix in displaying time if it is unavailable
+	
+	Revision 2.81  2001/01/11 06:49:58  crosser
+	minor cleanups on source
+	make manual page in sync with the source
+	
+	Revision 2.80  2000/08/13 21:51:48  crosser
+	More changes incorporated from John Bowman's version
+	
+	Revision 2.79  2000/07/14 06:35:39  crosser
+	reportedly some models return CAN instead of DC1
+	
+	Revision 2.78  2000/05/10 05:20:42  crosser
+	more of Bowman's Nikon things
+	
+	Revision 2.76  2000/05/02 22:26:35  crosser
+	A few things incorporated from John Bowman's Nikon specific diffs
+	
+	Revision 2.75  2000/03/14 21:15:35  crosser
+	Nikon cp950 seems to report negative file size, try negate it.
+	
 	Revision 2.74  2000/02/17 21:29:55  crosser
 	final cleanup for 3.04, change version
 	change debugging levels
@@ -358,10 +385,14 @@ static char *rcsid="$Id: photopc.c,v 2.74 2000/02/17 21:29:55 crosser Exp $";
 #define S_ISDIR(st_mode) ((S_IFDIR & (st_mode)) ? 1 : 0)
 #endif
 
-#ifdef UNIX
-static char *device="/dev/photopc";
+#ifdef DEFAULT_DEVICE
+static char *device=DEFAULT_DEVICE;
 #else
+# ifdef UNIX
+static char *device="/dev/photopc";
+# else
 static char *device="COM1:";
+# endif
 #endif
 static int debug=0;
 static int quiet=0;
@@ -390,6 +421,20 @@ file_abort_cleanup(void)
 	}
 }
 
+static long
+numphotos(eph_iob *iob)
+{
+	long r10,r40;
+
+	if (eph_getint(iob,40,&r40)) {
+		r40=-1;
+	}
+	if (eph_getint(iob,10,&r10)) {
+		r10=-1;
+	}
+	return (r10>r40)?r10:r40;
+}
+
 static int
 xchdir(eph_iob *iob,char *where)
 {
@@ -409,7 +454,7 @@ xchdir(eph_iob *iob,char *where)
 		}
 	}
 	for (p=strtok(path,"\\");p;p=strtok(NULL,"\\")) {
-		if (eph_setvar(iob,84,p,strlen(p))) {
+		if (eph_setvar(iob,84,p,(off_t)strlen(p))) {
 			fprintf(stderr,"could not chdir to \"%s\"\n",p);
 			return -1;
 		}
@@ -466,6 +511,7 @@ descend(eph_iob *iob,int depth,char *root,int(*each)(eph_iob *iob,char *path))
 		}
 		if (xchdir(iob,root)) {
 			fprintf(stderr,"restore dir to \"%s\" failed\n",root);
+			free(folder);
 			return -1;
 		}
 	}
@@ -477,12 +523,16 @@ descend(eph_iob *iob,int depth,char *root,int(*each)(eph_iob *iob,char *path))
 static int
 nonempty(eph_iob *iob,char *path)
 {
-	unsigned long result;
+	long result;
 
 	if (eph_getint(iob,10,&result)) {
 		return -1;
 	}
 	if (result) {
+		if (strlen(path) < 2) {
+			if (!quiet) fprintf(stderr,"Buggy camera: nonempty root folder \"%s\", skip over\n",path);
+			return 0;
+		}
 		if (!quiet) printf("Starting in folder \"%s\"\n",path);
 		return 1;
 	} else return 0;
@@ -523,8 +573,8 @@ setclock(eph_iob *iob,int argc,char *argv[])
 		now+=zoneoffset(now);
 	}
 
-	if (eph_setint(iob,2,now)) return -1;
-	if (eph_getint(iob,2,&new)) return -1;
+	if (eph_setint(iob,2,(long)now)) return -1;
+	if (eph_getint(iob,2,(long*)&new)) return -1;
 	if (labs(now-new) > 59L) {
 		fprintf(stderr,"time we tried to set does not match the result\n");
 		return -1;
@@ -556,25 +606,62 @@ int
 shutter(eph_iob *iob,int argc,char *argv[])
 {
 	long val=0;
+	int adjust=0;
+	char buffer[]={0,0,0,0,10,0,0,0};
 
 	if (strcasecmp(argv[1],"Auto") == 0) {
 		val=0L;
 	} else if ((strlen(argv[1]) > 2) && (strncmp(argv[1],"1/",2) == 0)) {
 		val=atol(argv[1]+2);
-		if (val <= 0) {
+		if (val == 0) {
 			fprintf(stderr,"bad shutter speed `%s'\n",argv[1]);
 			return -1;
 		} else {
 			val=1000000L/val;
 		}
-	} else {
-		if ((val=atol(argv[1])) <= 0) {
-			fprintf(stderr,"bad shutter speed `%s'\n",argv[1]);
-			return -1;
-		}
+	} else if (sscanf(argv[1],"plus%ld",&val) == 1) {
+		adjust=1;
+	} else if (sscanf(argv[1],"minus%ld",&val) == 1) {
+		adjust=-1;
+	} else if ((val=atol(argv[1])) == 0) {
+		fprintf(stderr,"bad shutter speed `%s'\n",argv[1]);
+		return -1;
 	}
 
-	return eph_setint(iob,3,val);
+	if (adjust) {
+		if (val == 0) {
+			fprintf(stderr,"bad shutter adjustment `%s'\n",argv[1]);
+			return -1;
+		} else {
+			val=val*adjust;
+		}
+		buffer[0]=(val>>0)&0xff;
+		buffer[1]=(val>>8)&0xff;
+		buffer[2]=(val>>16)&0xff;
+		buffer[3]=(val>>24)&0xff;
+		return eph_setvar(iob,69,buffer,sizeof(buffer));
+	} else return eph_setint(iob,3,val);
+}
+
+int
+aperture(eph_iob *iob,int argc,char *argv[])
+{
+	long val=0;
+
+	if (strcasecmp(argv[1],"Auto") == 0) {
+		val=0;
+	} else if (strcasecmp(argv[1],"Lo") == 0) {
+		val=1;
+	} else if (strcasecmp(argv[1],"Med") == 0) {
+		val=2;
+	} else if (strcasecmp(argv[1],"Hi") == 0) {
+		val=4;
+	} else if ((val=atol(argv[1])) == 0) {
+		fprintf(stderr,"bad aperture `%s'\n",argv[1]);
+		return -1;
+	}
+
+	return eph_setint(iob,5,val);
 }
 
 int
@@ -655,53 +742,161 @@ lcd_brightness(eph_iob *iob,int argc,char *argv[])
 int
 setid(eph_iob *iob,int argc,char *argv[])
 {
-	return eph_setvar(iob,0x16,argv[1],strlen(argv[1]));
+	return eph_setvar(iob,0x16,argv[1],(off_t)strlen(argv[1]));
 }
 
 int
-macro(eph_iob *iob,int argc,char *argv[])
+focus(eph_iob *iob,int argc,char *argv[])
 {
 	long val=0;
 
-	if (strcasecmp(argv[1],"On") == 0) {
+	if (strcasecmp(argv[1],"Macro") == 0) {
 		val=1;
-	} else if (strcasecmp(argv[1],"Off") == 0) {
+	} else if (strcasecmp(argv[1],"Normal") == 0) {
 		val=2;
-	} else {
-		fprintf(stderr,"bad macro mode `%s'\n",argv[1]);
+	} else if (strcasecmp(argv[1],"Infinity") == 0) {
+		val=3;
+	} else if ((val=atol(argv[1])) == 0) {
+		fprintf(stderr,"bad focus mode `%s'\n",argv[1]);
 		return -1;
 	}
 
 	return eph_setint(iob,33,val);
 }
 
+int
+adjust(eph_iob *iob,int argc,char *argv[])
+{
+	long val=0;
+
+	if (strcasecmp(argv[1],"Normal") == 0) {
+		val=0;
+	} else if (strcasecmp(argv[1],"Contrast+") == 0) {
+		val=1;
+	} else if (strcasecmp(argv[1],"Contrast-") == 0) {
+		val=2;
+	} else if (strcasecmp(argv[1],"Brightness+") == 0) {
+		val=3;
+	} else if (strcasecmp(argv[1],"Brightness-") == 0) {
+		val=4;
+	} else if ((val=atol(argv[1])) == 0) {
+		fprintf(stderr,"bad adjust mode `%s'\n",argv[1]);
+		return -1;
+	}
+
+	return eph_setint(iob,19,val);
+}
+
+int
+whitebalance(eph_iob *iob,int argc,char *argv[])
+{
+	long val=0;
+
+	if (strcasecmp(argv[1],"Auto") == 0) {
+		val=0;
+	} else if (strcasecmp(argv[1],"Sunny") == 0) {
+		val=1;
+	} else if (strcasecmp(argv[1],"Incandescent") == 0) {
+		val=2;
+	} else if (strcasecmp(argv[1],"Fluorescent") == 0) {
+		val=3;
+	} else if (strcasecmp(argv[1],"Flash") == 0) {
+		val=5;
+	} else if (strcasecmp(argv[1],"Preset") == 0) {
+		val=6;
+	} else if (strcasecmp(argv[1],"Cloudy") == 0) {
+		val=255;
+	} else if ((val=atol(argv[1])) == 0) {
+		fprintf(stderr,"bad whitebalance mode `%s'\n",argv[1]);
+		return -1;
+	}
+
+	return eph_setint(iob,20,val);
+}
+
+int
+metering(eph_iob *iob,int argc,char *argv[])
+{
+	long val=0;
+
+	if (strcasecmp(argv[1],"Center-weighted") == 0) {
+		val=2;
+	} else if (strcasecmp(argv[1],"Spot") == 0) {
+		val=3;
+	} else if (strcasecmp(argv[1],"Matrix") == 0) {
+		val=5;
+	} else if ((val=atol(argv[1])) == 0) {
+		fprintf(stderr,"bad metering mode `%s'\n",argv[1]);
+		return -1;
+	}
+
+	return eph_setint(iob,70,val);
+}
+
 int
 color(eph_iob *iob,int argc,char *argv[])
 {
 	long val=0;
 
-	if (strcasecmp(argv[1],"On") == 0) {
-		val=1;
-	} else if (strcasecmp(argv[1],"Off") == 0) {
-		val=2;
+	if (eph_getint(iob,72,&val)) {
+		if (strcasecmp(argv[1],"On") == 0) {
+			val=1;
+		} else if (strcasecmp(argv[1],"Off") == 0) {
+			val=2;
+		} else if ((val=atol(argv[1])) == 0) {
+			fprintf(stderr,"bad color mode `%s'\n",argv[1]);
+			return -1;
+		}
+		return eph_setint(iob,6,val);
 	} else {
-		fprintf(stderr,"bad color mode `%s'\n",argv[1]);
-		return -1;
+		if (strcasecmp(argv[1],"On") == 0) {
+			val&=~16;
+		} else if (strcasecmp(argv[1],"Off") == 0) {
+			val|=16;
+		} else {
+			fprintf(stderr,"bad color mode `%s'\n",argv[1]);
+			return -1;
+		}
+		return eph_setint(iob,72,val);
 	}
-
-	return eph_setint(iob,6,val);
 }
 
 int
 dzoom(eph_iob *iob,int argc,char *argv[])
 {
 	long val=0;
+	long highbits;
 
-	if (strcasecmp(argv[1],"On") == 0) {
-		val=8;
+	if (eph_getint(iob,72,&val)) return -1;
+	highbits=(val & 0x1F) | 8;
+
+	if (strcasecmp(argv[1],"Zero") == 0) {
+		val=0;
+	} else if (strcasecmp(argv[1],"AEL/WBL") == 0) {
+		val |= 1;
+	} else if (strcasecmp(argv[1],"Reset") == 0) {
+		val &= 0xFFFE;
+	} else if (strcasecmp(argv[1],"Fisheye") == 0) {
+		val=(val & 0xFFF1) | 2;
+	} else if (strcasecmp(argv[1],"Wide") == 0) {
+		val=(val & 0xFFF1) | 4;
+	} else if (strcasecmp(argv[1],"Telephoto") == 0) {
+		val=(val & 0xFFF9);
+	} else if (strcasecmp(argv[1],"Auto") == 0) {
+		val=highbits & 0x17;
+	} else if (strcasecmp(argv[1],"1.25X") == 0) {
+		val=highbits | 0x0100;
+	} else if (strcasecmp(argv[1],"1.6X") == 0) {
+		val=highbits | 0x0200;
+	} else if (strcasecmp(argv[1],"2.0X") == 0) {
+		val=highbits | 0x0300;
+	} else if (strcasecmp(argv[1],"2.5X") == 0) {
+		val=highbits | 0x0400;
+	} else if (strcasecmp(argv[1],"1.0X") == 0) {
+		val=highbits | 0x0500;
 	} else if (strcasecmp(argv[1],"Off") == 0) {
-		val=0; /* not sure if it is correct */
-	} else {
+		val=highbits | 0x0500;
+	} else if ((val=atol(argv[1])) == 0) {
 		fprintf(stderr,"bad dzoom mode `%s'\n",argv[1]);
 		return -1;
 	}
@@ -709,6 +904,20 @@ dzoom(eph_iob *iob,int argc,char *argv[])
 	return eph_setint(iob,72,val);
 }
 
+int
+zoom(eph_iob *iob,int argc,char *argv[])
+{
+	double val;
+	char buffer[]={0,0,0,0,10,0,0,0};
+
+	if ((val=atof(argv[1])) <= 0.0 || val >= 26.0) {
+		fprintf(stderr,"bad zoom value `%s'\n",argv[1]);
+		return -1;
+	}
+	buffer[0]=(unsigned int)(10.0*val-0.5);
+	return eph_setvar(iob,71,buffer,sizeof(buffer));
+}
+
 int
 seti(eph_iob *iob,int argc,char *argv[])
 {
@@ -736,7 +945,7 @@ setv(eph_iob *iob,int argc,char *argv[])
 		fprintf(stderr,"bad register value \"%s\"\n",argv[1]);
 		return -1;
 	}
-	return eph_setvar(iob,reg,argv[2],strlen(argv[2]));
+	return eph_setvar(iob,reg,argv[2],(off_t)strlen(argv[2]));
 }
 
 int
@@ -775,7 +984,7 @@ protect(eph_iob *iob,int argc,char *argv[])
 	} else if (strcasecmp(argv[2],"Off") == 0) {
 		pmode=0;
 	} else {
-		fprintf(stderr,"bad protect mode `%s'\n",argv[1]);
+		fprintf(stderr,"bad protect mode `%s'\n",argv[2]);
 		return -1;
 	}
 	if ((frame=atol(argv[1])) <= 0) {
@@ -797,6 +1006,41 @@ eraseall(eph_iob *iob,int argc,char *argv[])
 	else return 0;
 }
 
+int
+eraselast(eph_iob *iob,int argc,char *argv[])
+{
+	char zero=0;
+
+	if (eph_action(iob,0,&zero,1)) return -1;
+	else return 0;
+}
+
+int
+mode(eph_iob *iob,int argc,char *argv[])
+{
+	char val=0;
+
+	if (strcasecmp(argv[1],"Off") == 0) {
+		val=1;
+	} else if (strcasecmp(argv[1],"Record") == 0) {
+		val=2;
+	} else if (strcasecmp(argv[1],"Play") == 0) {
+		val=3;
+	} else if (strcasecmp(argv[1],"Thumbnail") == 0) {
+		val=6;
+	} else if (strcasecmp(argv[1],"Next") == 0) {
+		val=7;
+	} else if (strcasecmp(argv[1],"Previous") == 0) {
+		val=8;
+	} else if ((val=atol(argv[1])) == 0) {
+		fprintf(stderr,"bad mode `%s'\n",argv[1]);
+		return -1;
+	}
+
+	if (eph_action(iob,8,&val,1)) return -1;
+	else return 0;
+}
+
 int
 cmd(eph_iob *iob,int argc,char *argv[])
 {
@@ -804,10 +1048,10 @@ cmd(eph_iob *iob,int argc,char *argv[])
 	int intarg=-1;
 	char chararg;
 
-	if ((ccode=atol(argv[1])) <= 0) {
+	if (strspn(argv[1],"0123456789") != strlen(argv[1])) {
 		fprintf(stderr,"bad code value \"%s\"\n",argv[1]);
 		return -1;
-	}
+	} else ccode=atol(argv[1]);
 	if (strspn(argv[2],"0123456789") == strlen(argv[2]))
 		intarg=atol(argv[2]);
 	if ((intarg >= 0) && (intarg <= 255)) {
@@ -818,12 +1062,22 @@ cmd(eph_iob *iob,int argc,char *argv[])
 }
 
 static char *flashval[] = {"Auto","Force","Off","AntiRedeye","SlowSync"};
-static char *resval[] = {"-bad-","Low","High","Extended"};
+static char *resval[] = {"Unrecognized","Low","High","Extended"};
+static char *macroval[] = {"Unrecognized","Macro","Normal","Infinity"};
+static char *aperval[] = {"Auto","Low","Medium","High"};
+static char *adjval[] = {"Normal","Contrast+","Contrast-","Brightness+",
+				"Brightness-","Auto"};
+static char *wbval[] = {"Auto","Sunny","Incandescent","Fluorescent",
+				"Unrecognized","Flash","Preset"};
+static char *meterval[] = {"Unrecognized","Unrecognized","Center","Spot",
+				"Unrecognized","Matrix","Spot-AF-Area"};
+static char *operval[] = {"Unrecognized","Off","Record","Play",
+				"Unrecognized","Unrecognized","Thumbnail"};
 
 int
 query(eph_iob *iob,int argc,char *argv[])
 {
-	unsigned long result;
+	long result;
 	char *buffer;
 	off_t bufsize;
 	int rc;
@@ -834,11 +1088,11 @@ query(eph_iob *iob,int argc,char *argv[])
 		printf("Resolution: %lu - %s\n",(unsigned long)result,
 			(result < 4)?resval[result]:"Model-specific");
 	else if (rc == DC1)
-		printf("Resulution unavailable\n");
+		printf("Resolution unavailable\n");
 	else goto failure;
 
 	if ((rc=eph_getint(iob,2,&result)) == 0) {
-		if ((result == 0L) || (result == (unsigned long)-1L)) {
+		if ((result == 0L) || (result == -1L)) {
 			printf("Camera time: not set (%ld)\n",result);
 		} else {
 			printf("Camera time: %s\n",
@@ -854,31 +1108,38 @@ query(eph_iob *iob,int argc,char *argv[])
 	else if (rc == DC1)
 		printf("Shutter unavailable\n");
 	else goto failure;
-/*
+
 	if ((rc=eph_getint(iob,5,&result)) == 0)
-		printf("Reg 5: %lu\n",(unsigned long)result);
+		printf("Aperture: %lu - %s\n",(unsigned long)result,
+			(result < 4)?aperval[result]:"Unrecognized");
 	else if (rc == DC1)
-		printf("Reg 5 unavailable\n");
+		printf("Aperture unavailable\n");
 	else goto failure;
-*/
+
 	if ((rc=eph_getint(iob,6,&result)) == 0)
 		printf("Color mode: %lu - %s\n",result,
-		(result == 1)?"Color":(result == 2)?"B/W":"Bad value");
+		(result == 1)?"Color":(result == 2)?"B/W":"Unrecognized");
 	else if (rc == DC1)
 		printf("Color mode unavailable\n");
 	else goto failure;
 
 	if ((rc=eph_getint(iob,7,&result)) == 0)
 		printf("Flash: %lu - %s\n",(unsigned long)result,
-			(result < 5)?flashval[result]:"Bad value");
+			(result < 5)?flashval[result]:"Unrecognized");
 	else if (rc == DC1)
 		printf("Flash mode unavailable\n");
 	else goto failure;
 
 	if ((rc=eph_getint(iob,10,&result)) == 0)
-		printf("Frames taken: %lu\n",(unsigned long)result);
+		printf("Frames in folder: %lu\n",(unsigned long)result);
 	else if (rc == DC1)
-		printf("Frames taken unavailable\n");
+		printf("Frames in folder unavailable\n");
+	else goto failure;
+
+	if ((rc=eph_getint(iob,40,&result)) == 0)
+		printf("Total frames: %lu\n",(unsigned long)result);
+	else if (rc == DC1)
+		printf("Total frames unavailable\n");
 	else goto failure;
 
 	if ((rc=eph_getint(iob,11,&result)) == 0)
@@ -913,11 +1174,40 @@ query(eph_iob *iob,int argc,char *argv[])
 
 	if ((rc=eph_getint(iob,33,&result)) == 0)
 		printf("Macro mode: %lu - %s\n",result,
-		(result == 1)?"On":(result == 2)?"Off":"Bad value");
+			(result < 4)?macroval[result]:"Unrecognized");
 	else if (rc == DC1)
 		printf("Macro mode unavailable\n");
 	else goto failure;
 
+	if ((rc=eph_getint(iob,34,&result)) == 0)
+		printf("Operation mode: %lu - %s\n",result,
+			(result < 7)?operval[result]:"Unrecognized");
+	else if (rc == DC1)
+		printf("Operation mode unavailable\n");
+	else goto failure;
+
+	if ((rc=eph_getint(iob,19,&result)) == 0)
+		printf("Image adjust: %lu - %s\n",(unsigned long)result,
+			(result < 6)?adjval[result]:"Unrecognized");
+	else if (rc == DC1)
+		printf("Image adjust unavailable\n");
+	else goto failure;
+
+	if ((rc=eph_getint(iob,20,&result)) == 0)
+		printf("White balance: %lu - %s\n",(unsigned long)result,
+			(result < 7)?wbval[result]:
+			(result == 255)?"Cloudy":"Unrecognized");
+	else if (rc == DC1)
+		printf("White balance unavailable\n");
+	else goto failure;
+
+	if ((rc=eph_getint(iob,70,&result)) == 0)
+		printf("Metering mode: %lu - %s\n",(unsigned long)result,
+			(result < 7)?meterval[result]:"Unrecognized");
+	else if (rc == DC1)
+		printf("Metering mode unavailable\n");
+	else goto failure;
+
 	if ((rc=eph_getint(iob,35,&result)) == 0)
 		printf("LCD brightness: %lu (of 7)\n",(unsigned long)result);
 	else if (rc == DC1)
@@ -972,8 +1262,31 @@ query(eph_iob *iob,int argc,char *argv[])
 		printf("Manufacturer unavailable\n");
 	else goto failure;
 
+	bufsize=2048; buffer[0]='\0';
+	if ((rc=eph_getvar(iob,69,&buffer,&bufsize)) == 0) {
+		long adj=(long)(buffer[0]&0xff)
+			+((long)(buffer[1]&0xff)<<8)
+			+((long)(buffer[2]&0xff)<<16)
+			+((long)(buffer[3]&0xff)<<24);
+		printf("Shutter adjustment: %ld\n",adj);
+	} else if (rc == DC1)
+		printf("Shutter adjustment unavailable\n");
+	else goto failure;
+
+	bufsize=2048; buffer[0]='\0';
+	if ((rc=eph_getvar(iob,71,&buffer,&bufsize)) == 0) {
+		long ezoom=(long)(buffer[0]&0xff)
+			+((long)(buffer[1]&0xff)<<8)
+			+((long)(buffer[2]&0xff)<<16)
+			+((long)(buffer[3]&0xff)<<24);
+		printf("Effective zoom: %.1f\n",ezoom*.1);
+	} else if (rc == DC1)
+		printf("Effective zoom unavailable\n");
+	else goto failure;
+
 	if ((rc=eph_getint(iob,72,&result)) == 0)
-		printf("Digital zoom mode: %lu - %s\n",result,
+		printf("Digital zoom mode: %lu - %s\n",
+			(unsigned long)result,
 		(result <= 1)?"Off":"On");
 	else if (rc == DC1)
 		printf("Digital zoom mode unavailable\n");
@@ -990,13 +1303,7 @@ failure:
 int
 count(eph_iob *iob,int argc,char *argv[])
 {
-	unsigned long result;
-
-	if (eph_getint(iob,10,&result)) {
-		printf("-1\n");
-		return -1;
-	}
-	else printf("%lu\n",result);
+	printf("%ld\n",numphotos(iob));
 	return 0;
 }
 
@@ -1062,7 +1369,7 @@ camsum(eph_iob *iob,int argc,char *argv[])
 	}
 	for (i=0;i<32;i+=4) {
 		res=(long)(buf[i]&0xff)+((long)(buf[i+1]&0xff)<<8)+
-			((long)(buf[i+2]&0xff)<<16)+((long)(buf[i+2]&0xff)<<24);
+			((long)(buf[i+2]&0xff)<<16)+((long)(buf[i+3]&0xff)<<24);
 		printf("%d: %08lx\n",i/4,res);
 	}
 	free(buf);
@@ -1093,7 +1400,7 @@ picsum(eph_iob *iob,int argc,char *argv[])
 	}
 	for (i=0;i<32;i+=4) {
 		res=(long)(buf[i]&0xff)+((long)(buf[i+1]&0xff)<<8)+
-			((long)(buf[i+2]&0xff)<<16)+((long)(buf[i+2]&0xff)<<24);
+			((long)(buf[i+2]&0xff)<<16)+((long)(buf[i+3]&0xff)<<24);
 		printf("%d: %08lx ",i/4,res);
 	}
 	printf("\n");
@@ -1164,35 +1471,35 @@ listfiles(eph_iob *iob,char *path)
 	int rc;
 	long res,j,max;
 
-	if (eph_getint(iob,10,&max)) {
+	if ((max=numphotos(iob)) < 0) {
 		return -1;
 	}
 	if (have_folders) printf("%s: ",path);
 	if (max)
 		printf("total %lu\n"
-		"No.    Size R P Date and Time                Filename\n",
+		"No.    Size R P Date and Time                    Filename\n",
 			max);
 	else
 		printf("\n");
 
+	buffer=malloc(32);
+	bufsize=32;
 	for (j=1;j<=max;j++) {
 		printf("%3ld ",j);
 		if ((rc=eph_setint(iob,4,j)) == DC1) {
-			printf("N/A\n");
+			printf("%3ld N/A\n",j);
 			break;
 		} else if (rc) {
-			printf("FAIL\n");
+			printf("%3ld FAIL\n",j);
 			break;
 		}
 		if ((rc=eph_getint(iob,12,&res)) == DC1) {
-			printf("N/A    ");
+			printf("%3ld N/A    ",j);
 		} else if (rc) {
-			printf("FAIL   ");
+			printf("%3ld FAIL   ",j);
 		} else {
-			printf("%7lu ",res);
+			printf("%3ld %7ld ",j,res);
 		}
-		buffer=malloc(32);
-		bufsize=32;
 		if ((rc=eph_getvar(iob,47,&buffer,&bufsize)) == DC1) {
 			printf("N/A\n");
 			break;
@@ -1209,7 +1516,7 @@ listfiles(eph_iob *iob,char *path)
 		res=(long)(buffer[20]&0xff)+((long)(buffer[21]&0xff)<<8)+
 			((long)(buffer[22]&0xff)<<16)+((long)(buffer[23]&0xff)<<24);
 		if (res != -1L) printf("%8s",ctimetz(&res,usetimezone));
-		else printf("N/A");
+		else printf("N/A                         ");
 		if ((rc=eph_getvar(iob,79,&buffer,&bufsize)) == DC1) {
 			printf(" N/A\n");
 		} else if (rc) {
@@ -1232,7 +1539,7 @@ llistfiles(eph_iob *iob, char *path)
 	long j,max;
 	long res;
 
-	if (eph_getint(iob,10,&max)) {
+	if ((max=numphotos(iob)) < 0) {
 		return -1;
 	}
 	if (have_folders) printf("%s: ",path);
@@ -1264,7 +1571,7 @@ llistfiles(eph_iob *iob, char *path)
 		} else if (rc) {
 			printf("FAIL   ");
 		} else {
-			printf("%7lu ",res);
+			printf("%7ld ",res);
 		}
 		buffer=malloc(2048);
 		buffer[0]='\0';
@@ -1350,7 +1657,7 @@ running(off_t count)
 	if (!quiet && filesize)
 	{
 		printf("%lu: %lu of %lu\r",(unsigned long)frame,
-				(unsigned long)count,(unsigned long)filesize);
+				(unsigned long)count,filesize);
 		fflush(stdout);
 	}
 }
@@ -1384,7 +1691,7 @@ retrfile(eph_iob *iob,int argc,char *argv[],int datatype)
 #endif
 
 	buf8isize=32;
-	if ((buf8i=malloc(buf8isize)) == NULL) {
+	if ((buf8i=malloc((size_t)buf8isize)) == NULL) {
 		perror("malloc buf8i");
 		return -1;
 	}
@@ -1392,6 +1699,12 @@ retrfile(eph_iob *iob,int argc,char *argv[],int datatype)
 	filenm=argv[2];
 	if (eph_setint(iob,4,frame)) return -1;
 	if (eph_getint(iob,12,&ilength)) return -1;
+	if (ilength < 0) {
+		if (debug) fprintf(stderr,
+			"uh-oh, size of frame in reg12 negative (%ld),"
+			" will interpret as positive\n",ilength);
+		ilength=-ilength;
+	}
 	if (eph_setint(iob,4,frame)) return -1;
 	if (eph_getint(iob,13,&tlength)) return -1;
 	if (datatype == AUDIO) {
@@ -1421,7 +1734,7 @@ retrfile(eph_iob *iob,int argc,char *argv[],int datatype)
 		ilength2=(long)(buf8i[0]&0xff)+((long)(buf8i[1]&0xff)<<8)+
 				((long)(buf8i[2]&0xff)<<16);
 		if (ilength2 && (ilength == 0L)) {
-			if (!quiet) fprintf(stderr,
+			if (debug) fprintf(stderr,
 				"uh-oh, size of frame in reg12 (%ld) "
 				"is zero, will use size field from "
 				"picture summary register (%ld)!\n",
@@ -1494,7 +1807,7 @@ retrfile(eph_iob *iob,int argc,char *argv[],int datatype)
 		anyway.
 	*/
 	if (bufsize) {
-		buffer=malloc(bufsize);
+		buffer=malloc((size_t)bufsize);
 		if (buffer == NULL) {
 			fprintf(stderr,"could not alloc %lu bytes\n",
 					(unsigned long)bufsize);
@@ -1556,7 +1869,7 @@ retrfile(eph_iob *iob,int argc,char *argv[],int datatype)
 			int flen;
 
 			flen=strlen(filenm)+MAXFORMAT+5;
-			fname=malloc(flen);
+			fname=malloc((size_t)flen);
 			if (fname == NULL) {
 				if (buffer) free(buffer);
 				return -1;
@@ -1663,7 +1976,7 @@ preview(eph_iob *iob,int argc,char *argv[])
 	int rc=-1;
 
 	bufsize=2048;
-	buffer=malloc(bufsize);
+	buffer=malloc((size_t)bufsize);
 	if (buffer == NULL) {
 		fprintf(stderr,"could not alloc %lu bytes\n",
 				(unsigned long)bufsize);
@@ -1678,7 +1991,7 @@ preview(eph_iob *iob,int argc,char *argv[])
 	if (eph_action(iob,5,&zero,1)) goto exit;
 	if (eph_setint(iob,4,0L)) goto exit;
 	if (eph_getvar(iob,14,&buffer,&bufsize)) goto exit;
-	if (fwrite(buffer,bufsize,1,fp) != 1) {
+	if (fwrite(buffer,(size_t)bufsize,1,fp) != 1) {
 		perror(fname);
 		goto exit;
 	} else rc=0;
@@ -1700,7 +2013,7 @@ upload(eph_iob *iob,int argc,char *argv[])
 	FILE *fp=NULL;
 	char *buffer=NULL;
 	struct stat stbuf;
-	unsigned long res;
+	long res;
 	char zero=0;
 	int rc=-1;
 
@@ -1719,13 +2032,13 @@ upload(eph_iob *iob,int argc,char *argv[])
 				argv[1]);
 		goto failure;
 	}
-	buffer=(char*)malloc(stbuf.st_size);
+	buffer=(char*)malloc((size_t)stbuf.st_size);
 	if (buffer == NULL) {
 		fprintf(stderr,"upload cannot allocate %lu bytes: %s\n",
 				(unsigned long)stbuf.st_size,strerror(errno));
 		goto failure;
 	}
-	if (fread(buffer,stbuf.st_size,1,fp) != 1) {
+	if (fread(buffer,(size_t)stbuf.st_size,1,fp) != 1) {
 		fprintf(stderr,"upload cannot read from %s: %s\n",
 				argv[1],strerror(errno));
 		goto failure;
@@ -1795,10 +2108,10 @@ setbit(range_t *range,int index)
 		int newsize;
 
 		newsize=(index/128+1)*16;
-		tmp=(char *)malloc(newsize);
+		tmp=(char *)malloc((size_t)newsize);
 		if (tmp == NULL) return -1;
-		memset(tmp,0,newsize);
-		memcpy(tmp,range->bits,(range->max)/8);
+		memset(tmp,0,(size_t)newsize);
+		memcpy(tmp,range->bits,(size_t)(range->max)/8);
 		if (range->max) free(range->bits);
 		range->bits=tmp;
 		range->max=newsize*8;
@@ -1862,6 +2175,9 @@ iterate(int (*executor)(eph_iob *iob,int argc,char *argv[]),
 		if (strcasecmp(p,"all") == 0) {
 			bot=1;
 			top=-1;
+		} else if (strcasecmp(p,"last") == 0) {
+			bot=-1;
+			top=-1;
 		} else if ((strcasecmp(p+strlen(p)-5,"-last") == 0) &&
 			   (sscanf(p,"%ld-last",&bot) == 1)) {
 			top=-1;
@@ -1873,16 +2189,17 @@ iterate(int (*executor)(eph_iob *iob,int argc,char *argv[]),
 			fprintf(stderr,"bad frames interval \"%s\"\n",p);
 			return -1;
 		}
-		if (bot < 1) {
-			fprintf(stderr,"Bottom frame must be 1 or greater\n");
-			return -1;
-		}
 		if (top < 1) {
-			if (eph_getint(iob,0x0a,&top)) {
+			if ((top=numphotos(iob)) < 0) {
 				fprintf(stderr,"could not get number of frames\n");
 				return -1;
 			}
-			if (!quiet) printf("Top frame No. %ld\n",top);
+			if (bot < 1) bot=top;
+			if (!quiet) printf("Processing %ld-%ld\n",bot,top);
+			if (top == 0) {
+				if (!quiet) printf("No frames\n");
+				return 0;
+			}
 		}
 		if (top >= bot) step=1;
 		else step=-1;
@@ -1921,48 +2238,55 @@ struct _cmdlist {
 	int (*executor)(eph_iob *iob,int argc,char *argv[]);
 	char *help;
 } cmdlist[] = {
-	{"",		0,0,0,	NULL,		"== Query parameters =="	},
+	{"",		0,0,0,	NULL,		"== Query parameters =="},
 	{"query",	0,0,0,	query,		""			},
 	{"count",	0,0,0,	count,		""			},
 	{"list",	0,0,0,	list,		""			},
 	{"llist",	0,0,0,	llist,		""			},
+	{"cd",		1,0,0,	cd,		"<path>"		},
 	{"geti",	1,0,0,	geti,		"-<reg-no>"		},
 	{"getv",	1,0,0,	getv,		"-<reg-no>"		},
 	{"camsum",	0,0,0,	camsum,		"-"			},
-	{"picsum",	1,1,0,	picsum,		"-<frame-interval>"	},
-	{"allregs",	1,0,0,	allregs,	"-<max-reg-no>"			},
+	{"picsum",	1,1,0,	picsum,		"-<frame-list>"	},
+	{"allregs",	1,0,0,	allregs,	"-<max-reg-no>"		},
 	{"",		0,0,0,	NULL,		"== Do some actions =="	},
-	{"cd",		1,0,0,	cd,		"<path>"		},
-	{"erase",	1,-1,0,	erase,		"<frame-interval>"		},
-	{"protect",	2,1,0,	protect,	"`On' | `Off' <frame-interval>"	},
-	{"snapshot",	0,0,0,	snapshot,	""			},
+	{"protect",	2,1,0,	protect,	"<frame-list> `On' | `Off'"},
+	{"erase",	1,-1,0,	erase,		"<frame-list>"	},
 	{"eraseall",	0,0,0,	eraseall,	""			},
+	{"eraselast",	0,0,0,	eraselast,	""			},
+	{"snapshot",	0,0,0,	snapshot,	""			},
+	{"off",		0,0,0,	off,		""			},
 	{"",		0,0,0,	NULL,		"== Retreive files =="	},
-	{"image",	2,1,2,	image,		"`All' | <frame-interval> <filename>"	},
-	{"thumbnail",	2,1,2,	thumbnail,	"`All' | <frame-interval> <filename>"	},
-	{"audio",	2,1,2,	audio,		"`All' | <frame-interval> <filename>"	},
-	{"preview",	1,0,0,	preview,	"<filename>"	},
+	{"image",	2,1,2,	image,		"<frame-list> <path>"	},
+	{"thumbnail",	2,1,2,	thumbnail,	"<frame-list> <path>"	},
+	{"audio",	2,1,2,	audio,		"<frame-list> <path>"	},
+	{"preview",	1,0,0,	preview,	"<filename>"		},
 #ifndef LOWMEMORY
 	{"",		0,0,0,	NULL,		"== Upload files =="	},
 	{"upload",	1,0,0,	upload,		"<filename>"		},
 #endif
-	{"",		0,0,0,	NULL,		"== Set parameters =="			},
-	{"resolution",	1,0,0,	resolution,	"`Hi' | `Lo' | `Ext'"		},
+	{"",		0,0,0,	NULL,		"== Set parameters =="	},
+	{"mode",	1,0,0,	mode,		"`Record' | `Play' | `Thumbnail' | `Next' | `Previous'"},
+	{"resolution",	1,0,0,	resolution,	"`Hi' | `Lo' | `Ext' | <number>"},
 	{"clock",	0,0,0,	setclock,	""			},
-	{"shutter",	1,0,0,	shutter,	"`Auto' | <microseconds> | 1/<fraction-of-second>"			},
+	{"shutter",	1,0,0,	shutter,	"`Auto' | <microseconds> | 1/<fraction-of-second>"},
+	{"aperture",	1,0,0,	aperture,	"`Auto' | `Lo' | `Med' | `Hi'"},
 	{"flash",	1,0,0,	flash,		"`Auto' | `Force' | `Off' | `AntiRedeye' | `SlowSync'"},
 	{"id",		1,0,0,	setid,		"<string>"		},
 	{"autoshut-host",1,0,0,	autoshut_host,	"<seconds>"		},
 	{"autoshut-field",1,0,0,autoshut_field,	"<seconds>"		},
 	{"lcd-autoshut",1,0,0,	lcd_autoshut,	"<seconds>"		},
 	{"lcd-brightness",1,0,0,lcd_brightness,	"1 to 7"		},
-	{"macro",	1,0,0,	macro,		"`On' | `Off'"		},
+	{"focus",	1,0,0,	focus,		"`Macro' | `Normal' | `Infinity'"},
+	{"adjust",	1,0,0,	adjust,		"`Normal' | `Contrast+' | `Contrast-' | `Brightness+' | `Brightness-'"},
+	{"metering",	1,0,0,	metering,	"`Center-weighted' | `Spot' | `Matrix'"},
+	{"whitebalance",1,0,0,	whitebalance,	"`Normal' | `Sunny' | `Incandescent' | `Fluorescent' | `Flash' | `Preset' | `Cloudy'"		},
 	{"color",	1,0,0,	color,		"`On' | `Off'"		},
-	{"dzoom",	1,0,0,	dzoom,		"`On' | `Off'"		},
-	{"seti",	2,0,0,	seti,		"-<reg-no> <value-to-set>"	},
-	{"setv",	2,0,0,	setv,		"-<reg-no> <value-to-set>"	},
+	{"dzoom",	1,0,0,	dzoom,		"`On' | `Off' | <number>"},
+	{"zoom",	1,0,0,	zoom,		"<mm>"			},
+	{"seti",	2,0,0,	seti,		"-<reg-no> <value-to-set>"},
+	{"setv",	2,0,0,	setv,		"-<reg-no> <value-to-set>"},
 	{"cmd",		2,0,0,	cmd,		"-<cmd-code> <arg>"	},
-	{"off",		0,0,0,	off,		""			},
 	{"",		0,0,0,	NULL,		""			},
 	{NULL,		0,0,0,	NULL,		NULL			}
 };
@@ -1994,7 +2318,7 @@ showhelp(char *name)
 #endif
 	printf("\t-s baud\t- set communication speed to 9600,19200,38400,57600 or 115200\n");
 	printf("\t\t\tdefault is %d\n",MAX_SPEED);
-	printf("\t-S baud\t- set \"pseudo\" speed (see manual)\n");
+	printf("\t-S baud,baud\t- set \"pseudo\" speed (see manual)\n");
 	printf("\t-l dev\t- use device name instead of default %s\n\n",device);
 	printf("Commands:\n");
 	for (i=0;cmdlist[i].cmd;i++) {
@@ -2041,7 +2365,7 @@ showversion(char *name)
 			"normal"
 #endif
 		);
-	printf("     Copyright (c) 1997-1999 Eugene Crosser and Bruce Lightner\n");
+	printf("     Copyright (c) 1997-2001 Eugene Crosser and Bruce Lightner\n");
 	printf("     URL: http://www.average.org/digicam/\n");
 }
 
@@ -2226,7 +2550,7 @@ main(int argc,char *argv[])
 
 exit:
 /*
-	On older models, action 04 with zero paramater terminates
+	On older models, action 04 with zero parameter terminates
 	session but leaves the camera ON.  On Olympus 600 (at least)
 	this command not only terminates session but also turns the
 	camera off.  So, the "off" command will turn off newer models
